structure Reparser : REPARSER where type program = EL.program =
  struct
    type program = EL.program

    structure Lit = Literal

    exception Reparse of Pos.pos * string

    fun unsupported pos thing = raise Reparse (pos, thing ^ " not supported")

    fun reparse_path pos ids =
        let fun split ids =
            case ids of
                [] => raise Reparse (pos, "eep!  empty path?")
              | [id] => ([], id)
              | id::ids => let val (p, h) = split ids in (id::p, h) end
        in
            split ids
        end

    fun reparse_ty (ty, typos) =
        case ty of
            BigEL.TyCon ([], ["bool"]) => EL.TyBool
          | BigEL.TyCon ([], ["int"]) => EL.TyLit (Lit.LitTyInt)
          | BigEL.TyCon ([], ["char"]) => EL.TyLit (Lit.LitTyChar)
          | BigEL.TyCon ([], ["string"]) => EL.TyLit (Lit.LitTyString)
          | BigEL.TyCon ([], ["unit"]) => EL.TyUnit
          | BigEL.TyCon ([ty], ["list"]) =>
                let val ty = reparse_ty ty
                in
                    EL.TyList ty
                end
          | BigEL.TyCon (tys, path) =>
                let val tys = map reparse_ty tys
                    val longid = reparse_path typos path
                in
                    EL.TyCon (tys, longid)
                end
          | BigEL.TyVar tyv => EL.TyVar tyv
          | BigEL.TyRecord [("1", ty1), ("2", ty2)] =>
                let val ty1 = reparse_ty ty1
                    val ty2 = reparse_ty ty2
                in
                    EL.TyProd (ty1, ty2)
                end
          | BigEL.TyRecord _ => unsupported typos "arbitrary tuple/record types"
          | BigEL.TyArrow (ty1, ty2) =>
                let val ty1 = reparse_ty ty1
                    val ty2 = reparse_ty ty2
                in
                    EL.TyArrow (ty1, ty2)
                end

    fun flatten_pat (pat, patpos) =
        case pat of
            BigEL.PaVar [id] => id
            (*
            BigEL.PaVar [id] =>
                (id, NONE)
          | BigEL.PaConstrain ((BigEL.PaVar [id], _), ty) =>
                (id, SOME (reparse_ty ty))
            *)
          | _ => unsupported patpos "complex patterns"

    fun reparse_pair (epair, epairpos) =
        case epair of
            BigEL.ExRecord [("1", e1), ("2", e2)] =>
                let val e1 = reparse_exp e1
                    val e2 = reparse_exp e2
                in
                    (e1, e2)
                end
          | _ => raise Reparse (epairpos, "invalid primop args")

    and reparse_binopexp operator epair =
        let val (e1, e2) = reparse_pair epair
        in
            EL.ExBinOp (e1, operator, e2)
        end

    and reparse_pat_pair (ppair, ppairpos) =
        case ppair of
            BigEL.PaRecord [("1", p1), ("2", p2)] =>
                let val p1 = flatten_pat p1
                    val p2 = flatten_pat p2
                in
                    (p1, p2)
                end
          | _ => raise Reparse (ppairpos, "invalid primop args")

    and reparse_exp (exp, exppos) =
        case exp of
            BigEL.ExLit l => EL.ExLit l
          | BigEL.ExVar ["true"] => EL.ExBool true
          | BigEL.ExVar ["false"] => EL.ExBool false
          | BigEL.ExVar ["nil"] => EL.ExNil
          | BigEL.ExVar path =>
                let val longid = reparse_path exppos path
                in
                    EL.ExVar longid
                end
          | BigEL.ExFn [(p, e)] =>
                let val id = flatten_pat p
                    val e = reparse_exp e
                in
                    EL.ExFn (id, e)
                end
          | BigEL.ExFn _ => unsupported exppos "multi-clause fn expressions"
          | BigEL.ExApp ((BigEL.ExVar ["::"], _), epair) =>
                let val (e1, e2) = reparse_pair epair
                in
                    EL.ExCons (e1, e2)
                end
          | BigEL.ExApp ((BigEL.ExVar ["+"], _), epair) =>
                reparse_binopexp Lit.OpPlus epair
          | BigEL.ExApp ((BigEL.ExVar ["-"], _), epair) =>
                reparse_binopexp Lit.OpMinus epair
          | BigEL.ExApp ((BigEL.ExVar ["*"], _), epair) =>
                reparse_binopexp Lit.OpTimes epair
          | BigEL.ExApp ((BigEL.ExVar ["="], _), epair) =>
              reparse_binopexp Lit.OpEquals epair
          | BigEL.ExApp ((BigEL.ExVar ["^"], _), epair) =>
              reparse_binopexp Lit.OpConcat epair
          | BigEL.ExApp ((BigEL.ExVar["print"], _), earg) =>
              EL.ExProj2(EL.ExPair(EL.ExUnOp (Lit.OpPrint,reparse_exp earg),
                                   EL.ExUnit))

          | BigEL.ExApp ((BigEL.ExVar["int_to_string"], _), earg) =>
                EL.ExUnOp(Lit.OpIntToStr, reparse_exp earg)
          | BigEL.ExApp ((BigEL.ExVar["char_to_int"], _), earg) =>
                EL.ExUnOp(Lit.OpCharToInt, reparse_exp earg)
          | BigEL.ExApp ((BigEL.ExVar["chr_to_str"], _), earg) =>
                EL.ExUnOp(Lit.OpCharToInt, reparse_exp earg)
          | BigEL.ExApp ((BigEL.ExSelector "1", _), earg) =>
                let val earg = reparse_exp earg
                in
                    EL.ExProj1 earg
                end
          | BigEL.ExApp ((BigEL.ExSelector "2", _), earg) =>
                let val earg = reparse_exp earg
                in
                    EL.ExProj2 earg
                end
          | BigEL.ExApp (e1, e2) =>
                let val e1 = reparse_exp e1
                    val e2 = reparse_exp e2
                in
                    EL.ExApp (e1, e2)
                end
          | BigEL.ExCase (e, [((BigEL.PaVar ["nil"], _), enil),
                           ((BigEL.PaApp (["::"], ppair), _), econs)]) =>
                let val e = reparse_exp e
                    val (id1, id2) = reparse_pat_pair ppair
                    val enil = reparse_exp enil
                    val econs = reparse_exp econs
                in
                    EL.ExCase (e, enil, (id1, id2, econs))
                end
          | BigEL.ExCase _ => unsupported exppos "arbitrary case expressions"
          | BigEL.ExLet (decs, e) =>
                let val decs = reparse_decs decs
                    val e = reparse_exp e
                in
                    EL.ExLet (decs, e)
                end
          | BigEL.ExRecord [("1", e1), ("2", e2)] =>
                let val e1 = reparse_exp e1
                    val e2 = reparse_exp e2
                in
                    EL.ExPair (e1, e2)
                end
          | BigEL.ExRecord _ => unsupported exppos "arbitrary tuples/records"
          | BigEL.ExSelector _ => unsupported exppos "arbitrary record selectors"
          | BigEL.ExConstrain (e, ty) =>
                let val e = reparse_exp e
                    val ty = reparse_ty ty
                in
                    EL.ExConstrain (e, ty)
                end
          | BigEL.ExHandle _ => unsupported exppos "exceptions"
          | BigEL.ExRaise _ => unsupported exppos "exceptions"
          | BigEL.ExIf (e1, e2, e3) =>
                let val e1 = reparse_exp e1
                    val e2 = reparse_exp e2
                    val e3 = reparse_exp e3
                in
                    EL.ExIf (e1, e2, e3)
                end

    and reparse_dec (dec, decpos) =
        case dec of
            BigEL.DecVal (pat, exp) =>
                let val id = flatten_pat pat
                    val exp = reparse_exp exp
                in
                    EL.BndVal (id, exp)
                end
          | BigEL.DecFun [(funid, [argpat], NONE, exp)] =>
                let val argid = flatten_pat argpat
                    val exp = reparse_exp exp
                in
                    EL.BndFun (funid, argid, exp)
                end
          | BigEL.DecFun [(_, _, SOME _, _)] =>
                unsupported decpos "ascription on fun binding"
          | BigEL.DecFun [(_, _::_::_, _, _)] =>
                unsupported decpos "multi-arg fun bindings"
          | BigEL.DecFun _ =>
                unsupported decpos "mutually recursive fun bindings"
          | BigEL.DecType [(tyvars, id, ty)] =>
                let val ty = reparse_ty ty
                in
                    EL.BndType (tyvars, id, ty)
                end
          | BigEL.DecType _ => unsupported decpos "n-ary type declarations"
          | BigEL.DecDatatype _ => unsupported decpos "datatype declarations"
          | BigEL.DecException _ => unsupported decpos "exception declarations"
          | BigEL.DecLocal (decs1, decs2) =>
                let val bnds1 = reparse_decs decs1
                    val bnds2 = reparse_decs decs2
                in
                    EL.BndLocal (bnds1, bnds2)
                end
          | BigEL.DecOpen [path] =>
                let val longid = reparse_path decpos path
                in
                    EL.BndOpen longid
                end
          | BigEL.DecOpen _ => unsupported decpos "n-ary open declarations"
          | BigEL.DecStructure [(id, strexp)] =>
                let val strexp = reparse_strexp strexp
                in
                    EL.BndStructure (id, strexp)
                end
          | BigEL.DecStructure _ =>
                unsupported decpos "n-ary structure declarations"

    and reparse_decs decs = map reparse_dec decs

    and reparse_strexp (strexp, strexppos) =
        case strexp of
            BigEL.StrStruct strdecs =>
                let val bnds = reparse_strdecs strdecs
                in
                    EL.StrStruct bnds
                end
          | BigEL.StrVar path =>
                let val longid = reparse_path strexppos path
                in
                    EL.StrVar longid
                end
          | BigEL.StrSealOpaque (strexp, sigexp) =>
                let val strexp = reparse_strexp strexp
                    val sigexp = reparse_sigexp sigexp
                in
                    EL.StrSealOpaque (strexp, sigexp)
                end
          | BigEL.StrSealTransparent (strexp, sigexp) =>
                let val strexp = reparse_strexp strexp
                    val sigexp = reparse_sigexp sigexp
                in
                    EL.StrSealTransparent (strexp, sigexp)
                end
          | BigEL.StrApp _ => unsupported strexppos "functors"
          | BigEL.StrLet (strdecs, strexp) =>
                let val bnds = reparse_strdecs strdecs
                    val strexp = reparse_strexp strexp
                in
                    EL.StrLet (bnds, strexp)
                end

    and reparse_strdec (strdec, strdecpos) =
        case strdec of
            BigEL.SDecDec dec => reparse_dec dec
          | BigEL.SDecStructure x => reparse_dec (BigEL.DecStructure x, strdecpos)
          | BigEL.SDecLocal (strdecs1, strdecs2) =>
                let val bnds1 = reparse_strdecs strdecs1
                    val bnds2 = reparse_strdecs strdecs2
                in
                    EL.BndLocal (bnds1, bnds2)
                end

    and reparse_strdecs strdecs = map reparse_strdec strdecs

    and reparse_sigexp (sigexp, sigexppos) =
        case sigexp of
            BigEL.SigSig specs =>
                let val specs = reparse_specs specs
                in
                    EL.SigSig specs
                end
          | BigEL.SigVar _ => unsupported sigexppos "signature variables"
          | BigEL.SigWhere _ => unsupported sigexppos "where type = ..."

    and reparse_spec (spec, specpos) =
        case spec of
            BigEL.SpecVal (id, ty) =>
                let val ty = reparse_ty ty
                in
                    EL.SpecVal (id, ty)
                end
          | BigEL.SpecType [(tyvars, id, tyo)] =>
                let val tyo = Option.map reparse_ty tyo
                in
                    EL.SpecType (tyvars, id, tyo)
                end
          | BigEL.SpecType _ => unsupported specpos "n-ary type specifications"
          | BigEL.SpecDatatype _ => unsupported specpos "datatype specifications"
          | BigEL.SpecException _ => unsupported specpos "exception specifications"
          | BigEL.SpecStructure [(id, sigexp)] =>
                let val sigexp = reparse_sigexp sigexp
                in
                    EL.SpecStructure (id, sigexp)
                end
          | BigEL.SpecStructure _ =>
              unsupported specpos "n-ary structure specifications"
          | BigEL.SpecInclude _ => unsupported specpos "include specifications"
          | BigEL.SpecSharing _ => unsupported specpos "sharing specifications"

    and reparse_specs specs = map reparse_spec specs

    and reparse_topdec (topdec, topdecpos) =
        case topdec of
            BigEL.TDecStrDec strdec => reparse_strdec strdec
          | BigEL.TDecSignature _ => unsupported topdecpos "signature bindings"
          | BigEL.TDecFunctor _ => unsupported topdecpos "functor bindings"

    fun reparse_program topdecs = map reparse_topdec topdecs

    fun parse_string s = reparse_program (Parser.parseString s)
    fun parse_file file = reparse_program (Parser.parseFile file)
  end
