structure BigEL =
  struct
    type label = string
    type ident = string
    type tyvar = string

    type lit = Literal.lit

    type path = ident list (* length >= 0 *)

    datatype ty_ =
        TyCon of ty list * path
      | TyVar of tyvar
      | TyRecord of (label * ty) list
      | TyArrow of ty * ty

    withtype ty = ty_ * Pos.pos

    (* NB: usu. ident tybind; later, in SigWhere sigexps, we have path tybind *)
    type 'a tybind = tyvar list * 'a * ty
                    (* tyvarseq  tycon = ty *)
    type conbind = ident * ty option (* tycon [of ty] *)
    type datbind = tyvar list * ident * conbind list
                      (* tyvars tycon = conbinds *)

    datatype exp_ =
        ExLit of lit
      | ExVar of path
      | ExFn of (pat * exp) list
      | ExApp of exp * exp
      | ExCase of exp * (pat * exp) list
      | ExLet of dec list * exp
      | ExRecord of (label * exp) list
      | ExSelector of label (* #lab *)
      | ExConstrain of exp * ty
      | ExHandle of exp * (pat * exp) list
      | ExRaise of exp
      | ExIf of exp * exp * exp

    and pat_ =
        PaWild
      | PaVar of path
      | PaLit of lit
      | PaApp of path * pat
      | PaRecord of (label * pat) list
      | PaConstrain of pat * ty
      | PaAs of ident * pat

    and dec_ =
        DecVal of pat * exp
      | DecFun of (ident * pat list * ty option * exp) list
      | DecType of ident tybind list
      | DecDatatype of datbind list * ident tybind list
      | DecException of ident * ty option
      | DecLocal of dec list * dec list
      | DecOpen of path list
      | DecStructure of (ident * strexp) list
      (* no infix decs -- resolved at parse time *)

    and strexp_ =
        StrStruct of strdec list
      | StrVar of path
      | StrSealOpaque of strexp * sigexp (* strexp :> sigexp *)
      | StrSealTransparent of strexp * sigexp (* strexp : sigexp *)
      | StrApp of ident * strexp
      | StrLet of strdec list * strexp

    and sigexp_ =
        SigSig of spec list
      | SigVar of ident
      | SigWhere of sigexp * path tybind

    and strdec_ =
        SDecDec of dec
      | SDecStructure of (ident * strexp) list
      | SDecLocal of strdec list * strdec list

    and spec_ =
        SpecVal of ident * ty
      | SpecType of (tyvar list * ident * ty option) list
               (* type tyvarseq   tycon [= ty]  [and...] *)
      | SpecDatatype of datbind list (* datbind from DecDatatype above *)
      | SpecException of ident * ty option
      | SpecStructure of (ident * sigexp) list
      | SpecInclude of sigexp
      | SpecSharing of spec * path list (* list length >= 2 *)

    withtype exp = exp_ * Pos.pos
    and pat = pat_ * Pos.pos
    and dec = dec_ * Pos.pos

    and sigexp = sigexp_ * Pos.pos
    and strexp = strexp_ * Pos.pos
    and strdec = strdec_ * Pos.pos
    and spec = spec_ * Pos.pos

    datatype topdec_ =
        TDecStrDec of strdec
      | TDecSignature of (ident * sigexp) list
                 (* signature s = sig [and ...] *)
      | TDecFunctor of (ident * ident * sigexp * strexp) list
                 (* functor f   (m    : sig)   = str [and ...]*)

    withtype topdec = topdec_ * Pos.pos

    type program = topdec list


    (*** Derived forms ***)

    fun derivedProd con l =
        con (ListUtil.mapi (fn (c, i) => (Int.toString (i+1), c)) l)

    val tyProd = derivedProd TyRecord
    val paProd = derivedProd PaRecord
    val exProd = derivedProd ExRecord

    fun forget (x, pos) = x

    fun atdp f x = (f x, Pos.dummypos)
    fun atdp0 x = (x, Pos.dummypos)

    fun paList l =
        forget
          (foldr (fn (p, pp) => atdp PaApp (["::"], atdp paProd [p, pp]))
                 (atdp PaVar ["nil"])
                 l)

    fun exList l =
        forget
          (foldr (fn (e, ee) => atdp ExApp (atdp ExVar ["::"],
                                            atdp exProd [e, ee]))
                 (atdp ExVar ["nil"])
                 l)

    fun exSeq l =
        forget
          (ListUtil.combiner
            (fn (e, ee) => atdp ExCase (e, [(atdp0 PaWild, ee)])) l)

    fun exAndalso (e1, e2) = ExIf (e1, e2, atdp ExVar ["false"])
    fun exOrelse  (e1, e2) = ExIf (e1, atdp ExVar ["true"], e2)

    (* expand the datbinds of a datatype with respect to its withtype tybinds
       (c.f. the Definition) *)
    (* datbinds : (tyvar list * ident * (ident * ty option) list) list,
        (datatype ('a,...,'z)   tyid  =  Id      [ of ty ] | ...  and ...)

        tybinds : (tyvar list * ident * ty) list
        (withtype ('a,...,'z)   tyid  = ty  and ...) *)
    fun expand_datbinds (datbinds : datbind list, tybinds : ident tybind list) =
        let val tydefs = map (fn (tyvars, tyid, ty) => (tyid, (tyvars, ty)))
                             tybinds
            fun streq (x, y) =
                case String.compare (x, y) of
                    EQUAL => true
                  | _ => false

            fun subst env (ty, typos) =
                let val ty =
                    case ty of
                        TyCon (tys, path) =>
                            TyCon (map (subst env) tys, path)
                      | TyRecord labtys =>
                            TyRecord (map (fn (l, ty) => (l, subst env ty))
                                     labtys)
                      | TyArrow (ty1, ty2) =>
                            TyArrow (subst env ty1, subst env ty2)
                      | TyVar tyvar =>
                            (case ListUtil.Alist.find streq env tyvar of
                                NONE => ty
                              | SOME t => t)
                in
                    (ty, typos)
                end

            fun expand (ty, typos) =
                let val ty = 
                    case ty of
                        TyVar _ => ty
                      | TyRecord labtys =>
                            TyRecord (map (fn (lab, ty) => (lab, expand ty))
                                          labtys)
                      | TyArrow (ty1, ty2) =>
                            TyArrow (expand ty1, expand ty2)
                      | TyCon (tys, [tycon]) =>
                            let val tys = List.map expand tys
                            in
                                case ListUtil.Alist.find streq tydefs tycon of
                                    NONE => TyCon (tys, [tycon])
                                  | SOME (tyvars_i, ty_i) =>
                                        let val env =
                                                ListPair.map
                                                    (fn (tyv, (ty, _)) =>
                                                        (tyv, ty))
                                                    (tyvars_i, tys)
                                            val (t, _) = subst env ty_i
                                        in
                                            t
                                        end
                            end
                      | TyCon _ => ty
                in
                    (ty, typos)
                end

            val datbinds =
                map (fn (tyvars, tyid, conbinds) =>
                           (tyvars, tyid,
                            map (fn (constr, tyo) =>
                                    (constr, Option.map expand tyo))
                                conbinds))
                    datbinds
        in
            datbinds
        end
  end
