signature IL =
  sig
    exception Error of string

    (* labels *)

    (* NB: the various "label" constructors correspond roughly to the
       overbar function in HS *)
    datatype label =
        LabTycon of string
      | LabTyvar of string
      | LabExp of string
      | LabModule of string
      | LabInternal of string (* for public, private *)
      | LabStarred of label (* starred label *)

    val string_of_label : label -> string
    val label_eq : label * label -> bool

    type 'a var = 'a Variable.var

    type expvar = Variables.exp Variable.var
    type convar = Variables.con Variable.var
    type modvar = Variables.mod Variable.var

    structure ConVarSet : ORD_SET where type Key.ord_key = convar
    structure ModVarSet : ORD_SET where type Key.ord_key = modvar

    (* abstract syntax *)

    datatype kind =
        KndType
      | KndPi of convar * kind * kind (* convar bound in kind2 *)
      | KndProd of kind list
      | KndSing of con

    and con =
        ConVar of convar
      | ConEvar of ebind ref (* existential variable, for unification *)
      | ConAbs of convar * kind * con (* convar bound in con *)
      | ConApp of con * con
      | ConProd of con list
      | ConProj of int * con
      | ConModProj of module * label (* invariant: module is a value *)

      (* cons with kind KndType *)
      | TyBool
      | TyUnit
      | TyLit of Literal.lit_ty
      | TyProd of con * con
      | TyList of con
      | TyArrow of con * con
      | TyForall of convar * kind * con (* convar bound in con *)

    (* Evar binding *)
    and ebind =
        EvFree of ConVarSet.set * ModVarSet.set
               (* con deps,       module deps *)
               (* invariant: these variables are never
                  substituted into -- fatal error if they are *)
      | EvBound of con (* unified with con *)

    and exp =
        ExVar of expvar
      | ExUnit
      | ExBool of bool
      | ExIf of exp * exp * exp
      | ExLit of Literal.lit
      | ExUnOp of Literal.unop * exp
      | ExBinOp of exp * Literal.binop * exp
      | ExPair of exp * exp
      | ExProj1 of exp
      | ExProj2 of exp
      | ExNil of con (* nil[c] : c list *)
      | ExCons of exp * exp
      | ExCase of exp * exp * (expvar * expvar * exp) 
                (* expvar1, expvar2 bound in exp3 *)
      | ExFix of expvar * expvar * con * con * exp
                (* f      (x     : ty) : ty  = e *)
                (* expvar2 (x) bound in exp (e),
                   expvar1 (f) bound in all exps *)
      | ExApp of exp * exp
      | ExConAbs of convar * kind * exp (* convar bound in exp *)
      | ExConApp of exp * con
      | ExModLet of modvar * module * exp
      | ExModProj of module * label

    and module =
        ModVar of modvar
      | ModStruct of (label * bnd) list
      | ModModProj of module * label
      | ModConstrain of module * signat

    and bnd =
        BndExp of expvar * exp
      | BndCon of convar * con
      | BndMod of modvar * module
    (* NB: in a bnd list bnd::bnds, bnd's var bound in bnds *)

    and signat =
        SigStruct of (label * dec) list

    and dec =
        DecExpCon of expvar * con
      | DecConKind of convar * kind
      | DecModSig of modvar * signat
    (* NB: in a dec list decval ::decs, dec's var bound in decs *)

    type sdec = label * dec
    type sbnd = label * bnd

    (* derived forms *)
    val conMultiAbs : convar list * con -> con
    val tyMultiForall : convar list * con -> con

    val kndArrow : kind * kind -> kind
    val kndExp : kind * int -> kind
    val kndMultiPi : convar list * kind -> kind

    val exMultiConAbs : convar list * exp -> exp

    val labelledKndSing : con -> kind -> kind


    (* alpha variation *)

    (* [alpha_vary (var, phrase) replacefun] yields (var', phrase'), where
       var' is fresh, and (var. phrase) is alpha-equal to (var'. phrase') *)
    val alpha_vary : 'a var * 'b
                   -> ('a var -> 'a var -> 'b -> 'c) -> 'a var * 'c
    val alpha_vary2 : 'a var * 'b -> 'a var * 'b
                   -> ('a var -> 'a var -> 'b -> 'c) -> 'a var * ('c * 'c)


    (* various replacement functions suitable for use with the above *)

    (* by invariant, if called directly, these should only be called with
       a fresh second convar -- these do not perform alpha-conversion! *)
    val replace_con_con : convar -> convar -> con -> con
    val replace_con_exp : convar -> convar -> exp -> exp
    val replace_con_kind : convar -> convar -> kind -> kind
    val replace_con_module : convar -> convar -> module -> module
    val replace_con_sbnds : convar -> convar
                         -> (label * bnd) list -> (label * bnd) list
    val replace_con_sdecs : convar -> convar
                         -> (label * dec) list -> (label * dec) list
    val replace_con_signat : convar -> convar -> signat -> signat
    val replace_exp_con : expvar -> expvar -> con -> con
    val replace_exp_exp : expvar -> expvar -> exp -> exp
    val replace_exp_kind : expvar -> expvar -> kind -> kind
    val replace_exp_module : expvar -> expvar -> module -> module
    val replace_exp_sbnds : expvar -> expvar
                         -> (label * bnd) list -> (label * bnd) list
    val replace_exp_sdecs : expvar -> expvar
                         -> (label * dec) list -> (label * dec) list
    val replace_exp_signat : expvar -> expvar -> signat -> signat
    val replace_mod_con : modvar -> modvar -> con -> con
    val replace_mod_exp : modvar -> modvar -> exp -> exp
    val replace_mod_kind : modvar -> modvar -> kind -> kind
    val replace_mod_module : modvar -> modvar -> module -> module
    val replace_mod_sbnds : modvar -> modvar
                         -> (label * bnd) list -> (label * bnd) list
    val replace_mod_sdecs : modvar -> modvar
                         -> (label * dec) list -> (label * dec) list
    val replace_mod_signat : modvar -> modvar -> signat -> signat


    (* various substitution functions *)

    (* subst_xxx_phrase substitutes an xxx for an xxxvar in a phrase *)
    val subst_con_con : con -> convar -> con -> con
    val subst_con_exp : con -> convar -> exp -> exp
    val subst_con_kind : con -> convar -> kind -> kind
    val subst_con_module : con -> convar -> module -> module
    val subst_con_sbnds : con -> convar
                       -> (label * bnd) list -> (label * bnd) list
    val subst_con_sdecs : con -> convar
                       -> (label * dec) list -> (label * dec) list
    val subst_con_signat : con -> convar -> signat -> signat

    val subst_exp_con : exp -> expvar -> con -> con
    val subst_exp_exp : exp -> expvar -> exp -> exp
    val subst_exp_kind : exp -> expvar -> kind -> kind
    val subst_exp_module : exp -> expvar -> module -> module
    val subst_exp_sbnds : exp -> expvar
                       -> (label * bnd) list -> (label * bnd) list
    val subst_exp_sdecs : exp -> expvar
                       -> (label * dec) list -> (label * dec) list
    val subst_exp_signat : exp -> expvar -> signat -> signat

    val subst_mod_con : module -> ModVarSet.item -> con -> con
    val subst_mod_exp : module -> modvar -> exp -> exp
    val subst_mod_kind : module -> ModVarSet.item -> kind -> kind
    val subst_mod_module : module -> ModVarSet.item -> module -> module
    val subst_mod_sbnds : module -> modvar
                       -> (label * bnd) list -> (label * bnd) list
    val subst_mod_sdecs : module -> modvar
                       -> (label * dec) list -> (label * dec) list
    val subst_mod_signat : module -> modvar -> signat -> signat

    (* free modvars *)
    val fmv_bnd : bnd -> ModVarSet.set
    val fmv_con : con -> ModVarSet.set
    val fmv_dec : dec -> ModVarSet.set
    val fmv_exp : exp -> ModVarSet.set
    val fmv_kind : kind -> ModVarSet.set
    val fmv_mod : module -> ModVarSet.set
    val fmv_sig : signat -> ModVarSet.set

    (* various mapping functions for manipulating IL terms *)
    val map_con : (convar -> con)
               -> (ebind ref -> con)
               -> (convar * con -> convar * con)
               -> (kind -> kind)
               -> (module -> module)
               -> con -> con
    val map_exp : (expvar -> exp)
               -> (convar * exp -> convar * exp)
               -> (modvar * exp -> modvar * exp)
               -> (con -> con)
               -> (kind -> kind)
               -> (module -> module) -> exp -> exp
    val map_kind : (convar * kind -> convar * kind)
                -> (con -> con) -> kind -> kind
    val map_module : (modvar -> module)
                  -> ((label * bnd) list
                  -> (label * bnd) list)
                  -> (signat -> signat)
                  -> module -> module
    val map_sbnds : (expvar * ('a * bnd) list -> expvar * ('a * bnd) list)
                 -> (convar * ('a * bnd) list -> convar * ('a * bnd) list)
                 -> (modvar * ('a * bnd) list -> modvar * ('a * bnd) list)
                 -> (exp -> exp)
                 -> (con -> con)
                 -> (module -> module)
                 -> ('a * bnd) list -> ('a * bnd) list
    val map_sdecs : (expvar * ('a * dec) list -> expvar * ('a * dec) list)
                 -> (convar * ('a * dec) list -> convar * ('a * dec) list)
                 -> (modvar * ('a * dec) list -> modvar * ('a * dec) list)
                 -> (con -> con)
                 -> (kind -> kind)
                 -> (signat -> signat)
                 -> ('a * dec) list -> ('a * dec) list
    val map_signat : ((label * dec) list -> (label * dec) list)
                  -> signat -> signat
  end
