structure IL : IL =
  struct
    exception Error of string (* indicates internal fatal error *)

    type 'a var = 'a Variable.var

    (* the various "label" constructors correspond roughly to the overbar
       function in HS *)
    datatype label =
        LabTycon of string
      | LabTyvar of string
      | LabExp of string
      | LabModule of string
      | LabInternal of string (* for public, private *)
      | LabStarred of label (* starred label *)

    fun streq x y =
        case String.compare (x, y) of
            EQUAL => true
          | _ => false

    fun label_eq (lab1, lab2) =
        case (lab1, lab2) of
            (LabTycon s1, LabTycon s2) => streq s1 s2
          | (LabTyvar s1, LabTyvar s2) => streq s1 s2
          | (LabExp s1, LabExp s2) => streq s1 s2
          | (LabModule s1, LabModule s2) => streq s1 s2
          | (LabInternal s1, LabInternal s2) => streq s1 s2
          | (LabStarred l1, LabStarred l2) => label_eq (l1, l2)
          | _ => false

    fun string_of_label label =
        case label of
            LabTycon s => s
          | LabTyvar s => s
          | LabExp s => s
          | LabModule s => s
          | LabInternal s => s
          | LabStarred l => string_of_label l ^ "*"

    type expvar = Variables.exp Variable.var
    type convar = Variables.con Variable.var
    type modvar = Variables.mod Variable.var

    structure ConVarSet :> ORD_SET where type Key.ord_key = convar =
        SplaySetFn (struct
                        type ord_key = convar
                        val compare = Variable.compare
                    end)

    structure ModVarSet :> ORD_SET where type Key.ord_key = modvar =
        SplaySetFn (struct
                        type ord_key = modvar
                        val compare = Variable.compare
                    end)

    datatype kind =
        KndType
      | KndPi of convar * kind * kind (* convar bound in kind2 *)
      | KndProd of kind list
      | KndSing of con

    and con =
        ConVar of convar
      | ConEvar of ebind ref (* existential variable, for unification *)
      | ConAbs of convar * kind * con (* convar bound in con *)
      | ConApp of con * con
      | ConProd of con list
      | ConProj of int * con
      | ConModProj of module * label (* invariant: module is a value *)

      (* cons with kind KndType *)
      | TyBool
      | TyUnit
      | TyLit of Literal.lit_ty
      | TyProd of con * con
      | TyList of con
      | TyArrow of con * con
      | TyForall of convar * kind * con (* convar bound in con *)

    (* Evar binding *)
    and ebind =
        EvFree of ConVarSet.set * ModVarSet.set
               (* con deps,       module deps *)
               (* invariant: these variables are never
                  substituted into -- fatal error if they are *)
      | EvBound of con (* unified with con *)

    and exp =
        ExVar of expvar
      | ExUnit
      | ExBool of bool
      | ExIf of exp * exp * exp
      | ExLit of Literal.lit
      | ExUnOp of Literal.unop * exp
      | ExBinOp of exp * Literal.binop * exp
      | ExPair of exp * exp
      | ExProj1 of exp
      | ExProj2 of exp
      | ExNil of con (* nil[c] : c list *)
      | ExCons of exp * exp
      | ExCase of exp * exp * (expvar * expvar * exp) 
                (* expvar1, expvar2 bound in exp3 *)
      | ExFix of expvar * expvar * con * con * exp
                (* f      (x     : ty) : ty  = e *)
                (* expvar2 (x) bound in exp (e),
                   expvar1 (f) bound in all exps *)
      | ExApp of exp * exp
      | ExConAbs of convar * kind * exp (* convar bound in exp *)
      | ExConApp of exp * con
      | ExModLet of modvar * module * exp
      | ExModProj of module * label

    and module =
        ModVar of modvar
      | ModStruct of (label * bnd) list
      | ModModProj of module * label
      | ModConstrain of module * signat
      (* | ModLet of bnd list * module * signat *)
      (* XXX unnecessary? *)
        (* XXX or just modvar * module * module * signat ? *)

    and bnd =
        BndExp of expvar * exp
      | BndCon of convar * con
      | BndMod of modvar * module
    (* NB: in a bnd list bnd::bnds, bnd's var bound in bnds *)

    and signat =
        SigStruct of (label * dec) list

    and dec =
        DecExpCon of expvar * con
      | DecConKind of convar * kind
      | DecModSig of modvar * signat
    (* NB: in a dec list dec::decs, dec's var bound in decs *)

    type sdec = label * dec
    type sbnd = label * bnd

    fun labelledKndSing con knd =
        case knd of
            KndType => KndSing con
          | KndSing _ => KndSing con
          | KndPi (cx, k1, k2) =>
                KndPi (cx, k1, labelledKndSing (ConApp (con, ConVar cx)) k2)
          | KndProd knds =>
                KndProd
                   (ListUtil.mapi
                        (fn (k, i) => labelledKndSing (ConProj (i+1, con)) k)
                        knds)

    (* map_kind : (convar * kind -> convar * kind) -> (con -> con)
               -> kind -> kind  *)
    fun map_kind fconvarkind fcon knd =
        let val fkind = map_kind fconvarkind fcon
        in
            case knd of
                KndType => knd
              | KndProd knds => KndProd (map fkind knds)
              | KndSing c => KndSing (fcon c)
              | KndPi (x, k1, k2) =>
                    let val k1 = fkind k1
                        val (x, k2) = fconvarkind (x, k2)
                    in
                        KndPi (x, k1, k2)
                    end
        end

    (* map_con : (convar -> con) -> (ebind ref -> con)
              -> (convar * con -> convar * con)
              -> (kind -> kind) -> (module -> module) -> con -> con *)
    (* XXX put these args in a record?  and for other map_* functions ... *)
    fun map_con fconvar febindref fconvarcon fkind fmodule con =
            (* XXX jooc, better to do "fun fcon con = ... con"
                   or  "val fcon = ..." ?  optimization of partial appl.? *)
        let fun fcon con = map_con fconvar febindref fconvarcon fkind fmodule con
            fun fconvarkindcon (x, k, c) =
                    let val k = fkind k
                        val (x, c) = fconvarcon (x, c)
                    in
                        (x, k, c)
                    end
        in
            case con of
                ConVar x => fconvar x
              | ConEvar ebref => febindref ebref
              | ConAbs xkc => ConAbs (fconvarkindcon xkc)
              | ConApp (c1, c2) => ConApp (fcon c1, fcon c2)
              | ConProd cons => ConProd (map fcon cons)
              | ConProj (i, c) => ConProj (i, fcon c)
              | ConModProj (m, lab) => ConModProj (fmodule m, lab)
              | TyBool => TyBool
              | TyUnit => TyUnit
              | TyLit l => TyLit l
              | TyProd (c1, c2) => TyProd (fcon c1, fcon c2)
              | TyList c => TyList (fcon c)
              | TyArrow (c1, c2) => TyArrow (fcon c1, fcon c2)
              | TyForall xkc => TyForall (fconvarkindcon xkc)
        end

    (* map_module : (modvar -> module)
                 -> ((label * bnd) list -> (label * bnd) list)
                 -> (signat -> signat)
                 -> module -> module *)
    fun map_module fmodvar flabbnds fsignat module =
        let fun fmodule module = map_module fmodvar flabbnds fsignat module
        in
            case module of
                ModVar mv => fmodvar mv
              | ModStruct labbnds => ModStruct (flabbnds labbnds)
              | ModModProj (m, lab) => ModModProj (fmodule m, lab)
              | ModConstrain (m, sg) => ModConstrain (fmodule m, fsignat sg)
        end

    (* map_sbnds : (expvar * sbnds -> expvar * sbnds)
                -> (convar * sbnds -> convar * sbnds)
                -> (modvar * sbnds -> modvar * sbnds)
                -> (exp -> exp) -> (con -> con) -> (module -> module)
                -> sbnds -> sbnds
       (where sbnds = (label * bnd) list) *)
    fun map_sbnds fexpvarsbnds fconvarsbnds fmodvarsbnds
                  fexp fcon fmodule sbnds =
        case sbnds of
            [] => []
          | (lab, b)::sbs =>
                case b of
                    BndExp (ex, e) =>
                        let val e = fexp e
                            val (ex, sbs) = fexpvarsbnds (ex, sbs)
                        in
                            (lab, BndExp (ex, e)) :: sbs
                        end
                  | BndCon (cx, c) =>
                        let val c = fcon c
                            val (cx, sbs) = fconvarsbnds (cx, sbs)
                        in
                            (lab, BndCon (cx, c)) :: sbs
                        end
                  | BndMod (mx, m) =>
                        let val m = fmodule m
                            val (mx, sbs) = fmodvarsbnds (mx, sbs)
                        in
                            (lab, BndMod (mx, m)) :: sbs
                        end

    (* map_signat : ((lab * dec) list -> (lab * dec) list)
                 -> signat -> signat *)
    fun map_signat fsdecs signat =
        let fun fsignat signat = map_signat fsdecs signat
        in
            case signat of
                SigStruct sdecs => SigStruct (fsdecs sdecs)
        end

    (* map_sdecs : (expvar * sdecs -> expvar * sdecs)
                -> (convar * sdecs -> convar * sdecs)
                -> (modvar * sdecs -> modvar * sdecs)
                -> (con -> con) -> (kind -> kind) -> (signat -> signat)
                -> sdecs -> sdecs
       (where sdecs = (label * dec) list) *)
    fun map_sdecs fexpvarsdecs fconvarsdecs fmodvarsdecs
                  fcon fkind fsignat sdecs =
        case sdecs of
            [] => []
          | (lab, d)::sds =>
                case d of
                    DecExpCon (ex, c) =>
                        let val c = fcon c
                            val (ex, sds) = fexpvarsdecs (ex, sds)
                        in
                            (lab, DecExpCon (ex, c)) :: sds
                        end
                  | DecConKind (cx, k) =>
                        let val k = fkind k
                            val (cx, sds) = fconvarsdecs (cx, sds)
                        in
                            (lab, DecConKind (cx, k)) :: sds
                        end
                  | DecModSig (mx, sg) =>
                        let val sg = fsignat sg
                            val (mx, sds) = fmodvarsdecs (mx, sds)
                        in
                            (lab, DecModSig (mx, sg)) :: sds
                        end

    (* map_exp : (expvar -> exp)
              -> (convar * exp -> convar * exp)
              -> (modvar * exp -> modvar * exp)
              -> (con -> con)
              -> (kind -> kind)
              -> (module -> module)
              -> exp -> exp *)
    fun map_exp fexpvar
                fconvarexp
                fmodvarexp
                fcon
                fkind
                fmodule
                exp =
        let val fexp = map_exp fexpvar fconvarexp fmodvarexp fcon fkind fmodule
            fun flabexps labexps = map (fn (l, e) => (l, fexp e)) labexps
            fun fconvarkindexp (x, k, e) =
                let val k = fkind k
                    val (x, e) = fconvarexp (x, e)
                in
                    (x, k, e)
                end
            fun fmodvarmodexp (mx, m, e) =
                let val m = fmodule m
                    val (mx, e) = fmodvarexp (mx, e)
                in
                    (mx, m, e)
                end
        in
            case exp of
                ExVar x => fexpvar x
              | ExUnit => ExUnit
              | ExBool _ => exp
              | ExIf (e1, e2, e3) => ExIf (fexp e1, fexp e2, fexp e3)
              | ExLit _ => exp
              | ExUnOp (oper, e) => ExUnOp (oper, fexp e)
              | ExBinOp (e1, oper, e2) => ExBinOp (fexp e1, oper, fexp e2)
              | ExPair (e1, e2) => ExPair (fexp e1, fexp e2)
              | ExProj1 e => ExProj1 (fexp e)
              | ExProj2 e => ExProj2 (fexp e)
              | ExNil c => ExNil (fcon c)
              | ExCons (e1, e2) => ExCons (fexp e1, fexp e2)
              (* XXX should we do anything special with the bound vars in
                 ExCase, ExFix?  do we ever need to substitute into one of
                 these? *)
              | ExCase (e1, e2, (x, xs, e3)) =>
                    ExCase (fexp e1, fexp e2, (x, xs, fexp e3))
              | ExFix (f, x, con1, con2, e) =>
                    ExFix (f, x, fcon con1, fcon con2, fexp e)
              | ExApp (e1, e2) => ExApp (fexp e1, fexp e2)
              | ExConAbs xke => ExConAbs (fconvarkindexp xke)
              | ExConApp (e, c) => ExConApp (fexp e, fcon c)
              | ExModLet mx_m_e => ExModLet (fmodvarmodexp mx_m_e)
              | ExModProj (m, lab) => ExModProj (fmodule m, lab)
        end

    (* XXX are these correct? *)

    (* check_nocondep_ebindref : convar -> (con -> con) -> con *)
    (* [check_nocondep_ebindref convar k ebref] either
        (a) checks that convar doesn't appear in condeps
            if !ebref = EvFree (condeps, moddeps) or
        (b) returns [k con],
            if !ebref = EvBound con
        used in replace_con_con and subst_con_con *)
    fun check_nocondep_ebindref convar k ebref =
        case !ebref of
            EvFree (condeps, moddeps) =>
                if ConVarSet.member (condeps, convar) then
                    raise Error "substitution into Evar con dependency"
                else
                    ConEvar ebref
          | EvBound con => k con

    (* check_nomoddep_ebindref : modvar -> (con -> con) -> con *)
    (* same as above, but checks the module variable dependency set *)
    fun check_nomoddep_ebindref modvar k ebref =
        case !ebref of
            EvFree (condeps, moddeps) =>
                if ModVarSet.member (moddeps, modvar) then
                    raise Error "substitution into Evar module dependency"
                else
                    ConEvar ebref
          | EvBound con => k con


    (*** [replace_class1_class2 var var' c2] replaces all occurrences of
         var as a class1 object with var' inside a class2 object c2.
         this *doesn't* do any alpha conversion, so it's important that
         they only ever be called with a fresh var'.

         (these functions are used to implement alpha conversion, both in
         alpha equality and in capture-avoiding substitution, defined below
         as subst_class1_class2.)

         (c.f. alpha_vary, alpha_vary2)
    *)


    (** replacement of convars **)

    fun replace_con_kind convar convar' knd =
        map_kind (fn (cx, knd) =>
                    if Variable.eq (cx, convar) then
                        (* no free occurrences of var in knd -- bound by cx *)
                        (cx, knd)
                    else
                        (* cx not eq convar, so convar might be free in knd *)
                        (cx, replace_con_kind convar convar' knd))
                 (fn con => replace_con_con convar convar' con)
                 knd

    and replace_con_con convar convar' con =
        map_con (fn cx =>
                    if Variable.eq (cx, convar) then
                        ConVar convar'
                    else
                        ConVar cx)
                (check_nocondep_ebindref convar (replace_con_con convar convar'))
                (fn (cx, c) =>
                    if Variable.eq (cx, convar) then
                        (cx, c)
                    else
                        (cx, replace_con_con convar convar' c))
                (replace_con_kind convar convar')
                (replace_con_module convar convar')
                con

    and replace_con_module convar convar' module =
        map_module (fn mv => ModVar mv)
                   (replace_con_sbnds convar convar')
                   (replace_con_signat convar convar')
                   module


    and replace_con_sbnds convar convar' sbnds =
        map_sbnds (fn (ex, sbs) => (ex, replace_con_sbnds convar convar' sbs))
                  (fn (cx, sbs) => if Variable.eq (cx, convar) then
                                    (cx, sbs)
                                   else
                                    (cx, replace_con_sbnds convar convar' sbs))
                  (fn (mx, sbs) => (mx, replace_con_sbnds convar convar' sbs))
                  (replace_con_exp convar convar')
                  (replace_con_con convar convar')
                  (replace_con_module convar convar')
                  sbnds

    and replace_con_signat convar convar' signat =
        map_signat (replace_con_sdecs convar convar')
                   signat

    and replace_con_sdecs convar convar' sdecs =
        map_sdecs (fn (ex, sds) => (ex, replace_con_sdecs convar convar' sds))
                  (fn (cx, sds) => if Variable.eq (cx, convar) then
                                    (cx, sds)
                                   else
                                    (cx, replace_con_sdecs convar convar' sds))
                  (fn (mx, sds) => (mx, replace_con_sdecs convar convar' sds))
                  (replace_con_con convar convar')
                  (replace_con_kind convar convar')
                  (replace_con_signat convar convar')
                  sdecs

    and replace_con_exp convar convar' exp =
        map_exp (fn x => ExVar x)
                (fn (cx, e) =>
                    if Variable.eq (cx, convar) then
                        (cx, e)
                    else
                        (cx, replace_con_exp convar convar' e))
                (fn (mx, e) => (mx, replace_con_exp convar convar' e))
                (replace_con_con convar convar')
                (replace_con_kind convar convar')
                (replace_con_module convar convar')
                exp


    (** replacement of modvars **)

    fun replace_mod_kind modvar modvar' knd =
        map_kind (fn (cx, knd) => (cx, replace_mod_kind modvar modvar' knd))
                 (fn con => replace_mod_con modvar modvar' con)
                 knd

    and replace_mod_con modvar modvar' con =
        map_con (fn cx => ConVar cx)
                (check_nomoddep_ebindref modvar (replace_mod_con modvar modvar'))
                (fn (cx, c) => (cx, replace_mod_con modvar modvar' c))
                (replace_mod_kind modvar modvar')
                (replace_mod_module modvar modvar')
                con

    and replace_mod_module modvar modvar' module =
        map_module (fn mx =>
                        if Variable.eq (mx, modvar) then
                            ModVar modvar'
                        else
                            ModVar mx)
                   (replace_mod_sbnds modvar modvar')
                   (replace_mod_signat modvar modvar')
                   module


    and replace_mod_sbnds modvar modvar' sbnds =
        map_sbnds (fn (ex, sbs) => (ex, replace_mod_sbnds modvar modvar' sbs))
                  (fn (cx, sbs) => (cx, replace_mod_sbnds modvar modvar' sbs))
                  (fn (mx, sbs) => if Variable.eq (mx, modvar) then
                                    (mx, sbs)
                                   else
                                    (mx, replace_mod_sbnds modvar modvar' sbs))
                  (replace_mod_exp modvar modvar')
                  (replace_mod_con modvar modvar')
                  (replace_mod_module modvar modvar')
                  sbnds

    and replace_mod_signat modvar modvar' signat =
        map_signat (replace_mod_sdecs modvar modvar')
                   signat

    and replace_mod_sdecs modvar modvar' sdecs =
        map_sdecs (fn (ex, sds) => (ex, replace_mod_sdecs modvar modvar' sds))
                  (fn (cx, sds) => (cx, replace_mod_sdecs modvar modvar' sds))
                  (fn (mx, sds) => if Variable.eq (mx, modvar) then
                                    (mx, sds)
                                   else
                                    (mx, replace_mod_sdecs modvar modvar' sds))
                  (replace_mod_con modvar modvar')
                  (replace_mod_kind modvar modvar')
                  (replace_mod_signat modvar modvar')
                  sdecs

    and replace_mod_exp modvar modvar' exp =
        map_exp (fn x => ExVar x)
                (fn (cx, e) => (cx, replace_mod_exp modvar modvar' e))
                (fn (mx, e) =>
                    if Variable.eq (mx, modvar) then
                        (mx, e)
                    else
                        (mx, replace_mod_exp modvar modvar' e))
                (replace_mod_con modvar modvar')
                (replace_mod_kind modvar modvar')
                (replace_mod_module modvar modvar')
                exp


    (** replacement of expvars **)

    fun replace_exp_kind expvar expvar' knd =
        map_kind (fn (cx, knd) => (cx, replace_exp_kind expvar expvar' knd))
                 (fn con => replace_exp_con expvar expvar' con)
                 knd

    and replace_exp_con expvar expvar' con =
        map_con (fn cx => ConVar cx)
                (fn ebr => case !ebr of
                                EvFree _ => ConEvar ebr
                              | EvBound c => replace_exp_con expvar expvar' c)
                (fn (cx, c) => (cx, replace_exp_con expvar expvar' c))
                (replace_exp_kind expvar expvar')
                (replace_exp_module expvar expvar')
                con

    and replace_exp_module expvar expvar' module =
        map_module (fn mx => ModVar mx)
                   (replace_exp_sbnds expvar expvar')
                   (replace_exp_signat expvar expvar')
                   module


    and replace_exp_sbnds expvar expvar' sbnds =
        map_sbnds (fn (ex, sbs) => if Variable.eq (ex, expvar) then
                                    (ex, sbs)
                                   else
                                    (ex, replace_exp_sbnds expvar expvar' sbs))
                  (fn (cx, sbs) => (cx, replace_exp_sbnds expvar expvar' sbs))
                  (fn (mx, sbs) => (mx, replace_exp_sbnds expvar expvar' sbs))
                  (replace_exp_exp expvar expvar')
                  (replace_exp_con expvar expvar')
                  (replace_exp_module expvar expvar')
                  sbnds

    and replace_exp_signat expvar expvar' signat =
        map_signat (replace_exp_sdecs expvar expvar')
                   signat

    and replace_exp_sdecs expvar expvar' sdecs =
        map_sdecs (fn (ex, sds) => if Variable.eq (ex, expvar) then
                                    (ex, sds)
                                   else
                                    (ex, replace_exp_sdecs expvar expvar' sds))
                  (fn (cx, sds) => (cx, replace_exp_sdecs expvar expvar' sds))
                  (fn (mx, sds) => (mx, replace_exp_sdecs expvar expvar' sds))
                  (replace_exp_con expvar expvar')
                  (replace_exp_kind expvar expvar')
                  (replace_exp_signat expvar expvar')
                  sdecs

    and replace_exp_exp expvar expvar' exp =
        map_exp (fn x =>
                    if Variable.eq (x, expvar) then
                        ExVar expvar'
                    else
                        ExVar x)
                (fn (cx, e) => (cx, replace_exp_exp expvar expvar' e))
                (fn (mx, e) => (mx, replace_exp_exp expvar expvar' e))
                (replace_exp_con expvar expvar')
                (replace_exp_kind expvar expvar')
                (replace_exp_module expvar expvar')
                exp


    (*** [subst_class1_class1 class1 class1var class2] performs capture-
         avoiding substitution of class1 for class1var in class2 *)

    (* [alpha_vary (c1var, c2obj) replace] alpha-varies class1 bound
       variable c1var to a fresh variable in class2 object c2obj.
       replace replaces a class1 variable in a class2 object. *)
    fun alpha_vary (c1var, c2obj) replace =
        let val c1var' = Variable.freshen c1var
            val c2obj' = replace c1var c1var' c2obj
        in
            (c1var', c2obj')
        end

    (* alpha_vary2 : ('a var * 'b) -> ('a var * 'b)
                  -> ('a var -> 'a var -> 'b -> 'b)
                  -> ('a var * ('b * 'b))

       varies two binding constructs to use the same (fresh) variable.
       for computing alpha equality *)
    fun alpha_vary2 (varA, objA) (varB, objB) replace =
        let val var' = Variable.freshen varA
            val objA' = replace varA var' objA
            val objB' = replace varB var' objB
        in
            (var', (objA', objB'))
        end


    (* one thousand kinds of substitution follow.  be glad *you* didn't
       have to write this! *)

    (* substituting a con into a kind (cons appear in KndSing's) *)
    fun subst_con_kind con convar knd =
        map_kind (fn (cx, k) =>
                    if Variable.eq (cx, convar) then
                        (cx, k)
                    else
                        let (* eagerly alpha vary with a fresh variable *)
                            val (cx, k) = alpha_vary (cx, k) replace_con_kind
                            (* cx \not\in fv(con), so we won't incur capture *)
                            val k = subst_con_kind con convar k
                        in
                            (cx, k)
                        end)
                 (subst_con_con con convar)
                 knd

    and subst_con_con confor convar incon =
        map_con (fn cx =>
                    if Variable.eq (cx, convar) then
                        (* cx is convar free, substitute for it *)
                        confor
                    else
                        (* not convar -- put it back *)
                        ConVar cx)
                (check_nocondep_ebindref convar (subst_con_con confor convar))
                (fn (cx, c) =>
                    if Variable.eq (cx, convar) then
                        (* cx is convar being rebound, leave it alone *)
                        (cx, c)
                    else
                        (* different variable, eagerly rename and continue *)
                        let val (cx, c) = alpha_vary (cx, c) replace_con_con
                            (* now, cx \not\in fv(confor) *)
                            val c = subst_con_con confor convar c
                        in
                            (cx, c)
                        end)
                (subst_con_kind confor convar)
                (subst_con_module confor convar)
                incon

    and subst_con_module con convar module =
        map_module (fn mx => ModVar mx)
                   (subst_con_sbnds con convar)
                   (subst_con_signat con convar)
                   module

    and subst_con_sbnds con convar sbnds =
        map_sbnds (fn (ex, sbs) => (ex, subst_con_sbnds con convar sbs))
                  (fn (cx, sbs) => if Variable.eq (cx, convar) then   
                                    (cx, sbs)
                                   else
                                    let val (cx, sbs) = alpha_vary (cx, sbs)
                                                            replace_con_sbnds
                                        val sbs = subst_con_sbnds con convar sbs
                                    in
                                        (cx, sbs)
                                    end)
                  (fn (mx, sbs) => (mx, subst_con_sbnds con convar sbs))
                  (subst_con_exp con convar)
                  (subst_con_con con convar)
                  (subst_con_module con convar)
                  sbnds

    and subst_con_signat con convar signat =
        map_signat (subst_con_sdecs con convar)
                   signat

    and subst_con_sdecs con convar sdecs =
        map_sdecs (fn (ex, sds) => (ex, subst_con_sdecs con convar sds))
                  (fn (cx, sds) => if Variable.eq (cx, convar) then
                                    (cx, sds)
                                   else
                                    let val (cx, sds) = alpha_vary (cx, sds)
                                                            replace_con_sdecs
                                        val sds = subst_con_sdecs con convar sds
                                    in
                                        (cx, sds)
                                    end)
                  (fn (mx, sds) => (mx, subst_con_sdecs con convar sds))
                  (subst_con_con con convar)
                  (subst_con_kind con convar)
                  (subst_con_signat con convar)
                  sdecs

    and subst_con_exp con convar exp =
        map_exp (fn x => ExVar x)
                (fn (cx, e) =>
                    if Variable.eq (cx, convar) then
                        (cx, e)
                    else
                        let val (cx, e) = alpha_vary (cx, e) replace_con_exp
                            val e = subst_con_exp con convar e
                        in
                            (cx, e)
                        end)
                (fn (mx, e) => (mx, subst_con_exp con convar e))
                (subst_con_con con convar)
                (subst_con_kind con convar)
                (subst_con_module con convar)
                exp

    (* module substitution *)
    
    fun subst_mod_kind module modvar knd =
        map_kind (fn (cx, k) => (cx, subst_mod_kind module modvar k))
                 (subst_mod_con module modvar)
                 knd

    and subst_mod_con module modvar con =
        map_con (fn cx => ConVar cx)
                (check_nomoddep_ebindref modvar (subst_mod_con module modvar))
                (fn (cx, c) => (cx, subst_mod_con module modvar c))
                (subst_mod_kind module modvar)
                (subst_mod_module module modvar)
                con

    and subst_mod_module modulefor modvar modulein =
        map_module (fn mx =>
                        if Variable.eq (mx, modvar) then
                            modulefor
                        else
                            ModVar mx)
                   (subst_mod_sbnds modulefor modvar)
                   (subst_mod_signat modulefor modvar)
                   modulein

    and subst_mod_sbnds module modvar sbnds =
        map_sbnds (fn (ex, sbs) => (ex, subst_mod_sbnds module modvar sbs))
                  (fn (cx, sbs) => (cx, subst_mod_sbnds module modvar sbs))
                  (fn (mx, sbs) =>
                        if Variable.eq (mx, modvar) then   
                            (mx, sbs)
                        else
                            let val (mx, sbs) = alpha_vary (mx, sbs)
                                                           replace_mod_sbnds
                                val sbs = subst_mod_sbnds module modvar sbs
                            in
                                (mx, sbs)
                            end)
                  (subst_mod_exp module modvar)
                  (subst_mod_con module modvar)
                  (subst_mod_module module modvar)
                  sbnds

    and subst_mod_signat module modvar signat =
        map_signat (subst_mod_sdecs module modvar)
                   signat

    and subst_mod_sdecs module modvar sdecs =
        map_sdecs (fn (ex, sds) => (ex, subst_mod_sdecs module modvar sds))
                  (fn (cx, sds) => (cx, subst_mod_sdecs module modvar sds))
                  (fn (mx, sds) =>
                        if Variable.eq (mx, modvar) then
                            (mx, sds)
                        else
                            let val (mx, sds) = alpha_vary (mx, sds)
                                                           replace_mod_sdecs
                                val sds = subst_mod_sdecs module modvar sds
                            in
                                (mx, sds)
                            end)
                  (subst_mod_con module modvar)
                  (subst_mod_kind module modvar)
                  (subst_mod_signat module modvar)
                  sdecs

    and subst_mod_exp module modvar exp =
        map_exp (fn x => ExVar x)
                (fn (cx, e) => (cx, subst_mod_exp module modvar e))
                (fn (mx, e) =>
                    if Variable.eq (mx, modvar) then
                        (mx, e)
                    else
                        let val (mx, e) = alpha_vary (mx, e) replace_mod_exp
                            val e = subst_mod_exp module modvar e
                        in
                            (mx, e)
                        end)
                (subst_mod_con module modvar)
                (subst_mod_kind module modvar)
                (subst_mod_module module modvar)
                exp


    (* expression substitution *)

    fun subst_exp_kind exp expvar knd =
        map_kind (fn (cx, k) => (cx, subst_exp_kind exp expvar k))
                 (subst_exp_con exp expvar)
                 knd

    and subst_exp_con exp expvar con =
        map_con (fn cx => ConVar cx)
                (fn ebr => case !ebr of
                                EvFree _ => ConEvar ebr
                              | EvBound c => subst_exp_con exp expvar c)
                (fn (cx, c) => (cx, subst_exp_con exp expvar c))
                (subst_exp_kind exp expvar)
                (subst_exp_module exp expvar)
                con

    and subst_exp_module exp expvar module =
        map_module (fn mx => ModVar mx)
                   (subst_exp_sbnds exp expvar)
                   (subst_exp_signat exp expvar)
                   module

    and subst_exp_sbnds exp expvar sbnds =
        map_sbnds (fn (ex, sbs) => if Variable.eq (ex, expvar) then   
                                    (ex, sbs)
                                   else
                                    let val (ex, sbs) = alpha_vary (ex, sbs)
                                                            replace_exp_sbnds
                                        val sbs = subst_exp_sbnds exp expvar sbs
                                    in
                                        (ex, sbs)
                                    end)
                  (fn (cx, sbs) => (cx, subst_exp_sbnds exp expvar sbs))
                  (fn (mx, sbs) => (mx, subst_exp_sbnds exp expvar sbs))
                  (subst_exp_exp exp expvar)
                  (subst_exp_con exp expvar)
                  (subst_exp_module exp expvar)
                  sbnds

    and subst_exp_signat exp expvar signat =
        map_signat (subst_exp_sdecs exp expvar)
                   signat

    and subst_exp_sdecs exp expvar sdecs =
        map_sdecs (fn (ex, sds) => if Variable.eq (ex, expvar) then
                                    (ex, sds)
                                   else
                                    let val (ex, sds) = alpha_vary (ex, sds)
                                                            replace_exp_sdecs
                                        val sds = subst_exp_sdecs exp expvar sds
                                    in
                                        (ex, sds)
                                    end)
                  (fn (cx, sds) => (cx, subst_exp_sdecs exp expvar sds))
                  (fn (mx, sds) => (mx, subst_exp_sdecs exp expvar sds))
                  (subst_exp_con exp expvar)
                  (subst_exp_kind exp expvar)
                  (subst_exp_signat exp expvar)
                  sdecs

    and subst_exp_exp expfor expvar expin =
        map_exp (fn x =>
                    if Variable.eq (x, expvar) then
                        expfor
                    else
                        ExVar x)
                (fn (cx, e) => (cx, subst_exp_exp expfor expvar e))
                (fn (mx, e) => (mx, subst_exp_exp expfor expvar e))
                (subst_exp_con expfor expvar)
                (subst_exp_kind expfor expvar)
                (subst_exp_module expfor expvar)
                expin

    
    (* XXX fix these later/as needed -wjl *)

    (*** derived forms ***)

    val kndUnit = KndProd []
    val conUnit = ConProd []


    fun kndArrow (k1, k2) =
        let val cx = Variable.newNamed "_"
        in
            KndPi (cx, k1, k2)
        end

    fun kndExp (k, n) = KndProd (List.tabulate (n, fn _ => k))

    (* free module variables of a constructor *)
    fun fmv_con con =
        case con of
            ConVar _ => ModVarSet.empty
          | ConEvar ebr => ModVarSet.empty (* XXX should this be its moddeps? *)
          | ConAbs (cx, k, c) => fmv_con c
          | ConApp (c1, c2) => ModVarSet.union (fmv_con c1, fmv_con c2)
          | ConProd cs =>
                foldl (fn (c, set) => ModVarSet.union (fmv_con c, set))
                      ModVarSet.empty
                      cs
          | ConProj (i, c) => fmv_con c
          | ConModProj (module, lab) => fmv_mod module
          | TyBool => ModVarSet.empty
          | TyUnit => ModVarSet.empty
          | TyLit _ => ModVarSet.empty
          | TyProd (c1, c2) => ModVarSet.union (fmv_con c1, fmv_con c2)
          | TyList c => fmv_con c
          | TyArrow (c1, c2) => ModVarSet.union (fmv_con c1, fmv_con c2)
          | TyForall (cx, k, c) => ModVarSet.union (fmv_kind k, fmv_con c)

    and fmv_kind knd =
        case knd of
            KndType => ModVarSet.empty
          | KndPi (cx, k1, k2) => ModVarSet.union (fmv_kind k1, fmv_kind k2)
          | KndProd ks =>
                foldl (fn (k, set) => ModVarSet.union (fmv_kind k, set))
                      ModVarSet.empty
                      ks
          | KndSing c => fmv_con c

    and fmv_mod module =
        case module of
            ModVar mx => ModVarSet.singleton mx
          | ModStruct labbnds =>
                foldl (fn ((lab, bnd), set) =>
                        ModVarSet.union (fmv_bnd bnd, set))
                      ModVarSet.empty
                      labbnds
          | ModModProj (m, lab) => fmv_mod m
          | ModConstrain (m, sg) => ModVarSet.union (fmv_mod m, fmv_sig sg)

    and fmv_bnd bnd = 
        case bnd of
            BndExp (ex, e) => fmv_exp e
          | BndCon (cx, c) => fmv_con c
          | BndMod (mx, m) => ModVarSet.delete (fmv_mod m, mx)

    and fmv_sig signat =
        case signat of
            SigStruct sdecs =>
                foldl (fn ((lab, dec), set) =>
                        ModVarSet.union (fmv_dec dec, set))
                      ModVarSet.empty
                      sdecs

    and fmv_dec dec =
        case dec of
            DecExpCon (ex, c) => fmv_con c
          | DecConKind (cx, k) => fmv_kind k
          | DecModSig (mx, sg) => ModVarSet.delete (fmv_sig sg, mx)
    

    and fmv_exp exp =
        case exp of
            ExVar _ => ModVarSet.empty
          | ExUnit => ModVarSet.empty
          | ExBool _ => ModVarSet.empty
          | ExIf (e1, e2, e3) =>
                ModVarSet.union (fmv_exp e1,
                                 ModVarSet.union (fmv_exp e2, fmv_exp e3))
          | ExLit _ => ModVarSet.empty
          | ExUnOp (oper, e) => (fmv_exp e)
          | ExBinOp (e1, oper, e2) => ModVarSet.union (fmv_exp e1, fmv_exp e2)
          | ExPair (e1, e2) => ModVarSet.union (fmv_exp e1, fmv_exp e2)
          | ExProj1 e => fmv_exp e
          | ExProj2 e => fmv_exp e
          | ExNil c => fmv_con c
          | ExCons (e1, e2) => ModVarSet.union (fmv_exp e1, fmv_exp e2)
          | ExCase (e1, e2, (x, xs, e3)) =>
                ModVarSet.union (fmv_exp e1,
                                 ModVarSet.union (fmv_exp e2, fmv_exp e3))
          | ExFix (f, x, c1, c2, e) =>
                ModVarSet.union (fmv_con c1,
                                 ModVarSet.union (fmv_con c2, fmv_exp e))
          | ExApp (e1, e2) => ModVarSet.union (fmv_exp e1, fmv_exp e2)
          | ExConAbs (cx, k, e) => ModVarSet.union (fmv_kind k, fmv_exp e)
          | ExConApp (e, c) => ModVarSet.union (fmv_exp e, fmv_con c)
          | ExModLet (mx, m, e) =>
                ModVarSet.union (fmv_mod m,
                                 ModVarSet.delete (fmv_exp e, mx))
          | ExModProj (m, lab) => fmv_mod m

    (*
    (* con is the result type *)
    fun exIf con (e1, e2, e3) =
        let val ex = Variable.new ()
        in
            ExCase (tyBool, e1, [exLam (ex, tyUnit, con, e2),
                                 exLam (ex, tyUnit, con, e3)])
        end

    fun exPproj (labi, sumcon, exp, exn_exp) =
        let val labcons = case sumcon of
                            TySum lc => lc
                          | _ => raise Error "exPproj at non-sum type"
            val coni = case ListUtil.Alist.find label_eq labcons labi of
                            SOME con => con
                          | NONE => raise Error "exPproj on non-existent label"
            val ex : expvar = Variable.new ()
            val raise_exp = ExRaise (coni, exn_exp)
            val es = map (fn (lab, con) =>
                            exLam (ex, con, coni, if label_eq (lab, labi) then
                                                    ExVar ex
                                                  else
                                                    raise_exp))
                         labcons
        in
            ExCase (coni, exp, es)
        end
    *)

    (* (a1 : T, ..., an : T) . c; for use in conMultiAbs and conMultiMu below *)
    fun multiBind (convars : convar list, phrase) subst_con_phrase =
        let val cx : convar = Variable.new ()
            val n = List.length convars
            val k = kndExp (KndType, n)
            (* substitution = [pi_i cx / a1, ..., pi_n cx / an] (i = 1..n) *)
            val substitution =
                ListUtil.mapi
                          (fn (cxi, i) => (ConProj (i+1, ConVar cx), cxi))
                          convars
            val phrase =
                foldr (fn ((cproji, cxi), c) => subst_con_phrase cproji cxi c)
                      phrase
                      substitution
        in
            (cx, k, phrase)
        end

    (* \ (a1 : T, ..., an : T) . c *)
    fun conMultiAbs (convars, con) =
        let val cx_k_con = multiBind (convars, con) subst_con_con
        in
            ConAbs cx_k_con
        end

    (* /\ (a1 : T, ..., an : T) . e *)
    fun exMultiConAbs (convars, exp) =
        let val cx_k_exp = multiBind (convars, exp) subst_con_exp
        in
            ExConAbs cx_k_exp
        end

    fun tyMultiForall (convars, con) =
        let val cx_k_con = multiBind (convars, con) subst_con_con
        in
            TyForall cx_k_con
        end

    fun kndMultiPi (convars, knd) =
        let val cx_k_knd = multiBind (convars, knd) subst_con_kind
        in
            KndPi cx_k_knd
        end
  end
