structure Parser :> PARSER =
  struct
    open Parsing
    open Tokens
    open Literal
    open BigEL

    infixr 4 << >>
    infixr 3 &&
    infix  2 -- ##
    infix  2 wth suchthat return guard when
    infixr 1 ||

    exception Internal of string
    exception Parse of Pos.pos * string
    fun error str = fn pos => raise Parse (pos, str)

    (* parse something and check that the predicate holds, with an error o/w *)
    infix 2 check
    fun p check (pred, err) =
        !!p wth (fn (x, pos) => if pred x then x else raise Parse (pos, err x))

    fun streq x y =
        case String.compare (x, y) of
            EQUAL => true
          | _ => false

    (* val ` = literal *)
    fun `token = satisfy (fn x => eq (x, token))

    fun ifmany f l = case l of [x] => x
                             | _ => f l

    (* get a marked thing's position *)
    fun pos (x, p) = p
    (* forget a marked thing's position *)
    fun forget (x, p) = x
    (* lift a binary function to return a marked thing *)
    fun mark f (xp, yp) = (f (xp, yp), Pos.union (pos xp, pos yp))
    (* lift a 2-item-list function to return a marked thing *)
    fun markl2 f l =
        case l of
            [xp, yp] => (f l, Pos.union (pos xp, pos yp))
          | _ => raise Internal "invalid argument to markl2"
    (* like ifmany, but ensures the result is unmarked *)
    fun ifmanymark f l = case l of [x] => forget x
                                 | _ => f l

    (* lifts Option.map to carry through a pos *)
    fun option_mapmark f opt =
        case opt of
            NONE => NONE
          | SOME (xp as (x, pos)) => SOME (f xp, pos)

    (*** Various identifiers, short and long ***)

    val parseId    = maybe (fn IDENT s => SOME s | _ => NONE)
    val parseTyVar = maybe (fn TYVAR s => SOME s | _ => NONE)
    val parseLab   = maybe (fn IDENT s => SOME s
                             | INT i => SOME (Int.toString i)
                             | _ => NONE)

    (* structure identifiers must be alphanumeric, Definition p. 5 *)
    val parseStrId = parseId suchthat (fn s => Char.isAlpha (String.sub (s, 0)))
    (* "*" cannot be used as a type constructor, Definition p. 5 *)
    val parseTyId = parseId suchthat (fn s => not (streq s "*"))
    val parseStar = parseId suchthat (fn s => streq s "*")

    (* including trivial paths *)
    val parseTyPath = repeat (parseStrId << `DOT) && parseTyId
                        wth (fn (l, x) => l @ [x])
    val parseStrPath = parseStrId && repeat (`DOT >> parseStrId) wth op::

    fun uniqueStrings l =
        let val sorted = ListUtil.sort String.compare l
            fun unique l =
                case l of
                    nil => true
                  | [x] => true
                  | x1::(tl as x2::xs) => not (streq x1 x2) andalso unique tl
        in
            unique sorted
        end

    fun uniqueLabels l =
        let val labels = map (fn (x, _) => x) l
        in
            uniqueStrings labels
        end



    (*** Types ***)

    fun parseAtomicTy_ () =
        alt [parseTyVar wth TyVar,
             `LPAREN >> $parseTy_ << `RPAREN,
             `LBRACE >> separate0 (parseLab && `COLON >> $parseTy) (`COMMA)
                << `RBRACE
                check (uniqueLabels, fn _ => "duplicate record label in type")
                wth TyRecord]

    and parseAtomicTy () = !! ($parseAtomicTy_)

    and parseAppTy () =
    !! (
        alt [((`LPAREN >> separate ($parseTy) (`COMMA) << `RPAREN) || succeed [])
                && parseTyPath
                wth TyCon,
             $parseAtomicTy_]
       )
        &&
        repeat (!!parseTyPath)
        wth (fn (arg, tycons) =>
                foldl (mark (fn (tycon, a) => TyCon ([a], forget tycon)))
                      arg
                      tycons)
    and parseProdTy_ () = separate ($parseAppTy) parseStar wth ifmanymark tyProd

    and parseProdTy () = !! ($parseProdTy_)

    and parseArrowTy_ () =
        $parseProdTy && opt (`ARROW >> $parseTy)
            wth (fn (ty1, NONE) => forget ty1
                  | (ty1, SOME ty2) => TyArrow (ty1, ty2))

    and parseTy_ () = $parseArrowTy_
    and parseTy () = !! ($parseTy_)



    (*** Patterns ***)

    fun literal_of_token t =
        case t of
            INT n => SOME (LitInt n)
          | CHAR c => SOME (LitChar c)
          | STRING s => SOME (LitString s)
          | _ => NONE

    val parseLit = maybe literal_of_token


    (** in what follows, ctx is a FixContext.context,
        initial fixity context is FixContext.initial *)

    (* call a parser with a fixity context -- to prevent infinite recursion *)
    fun withCtx ctx parser = $(fn () => parser ctx)

    fun isNonfix ctx ident = not (FixContext.isInfix ctx ident)
    fun parseNonfixId ctx = parseId suchthat (isNonfix ctx) || `OP >> parseId

    (* take fixity into account for bare variables (trivial paths) *)
    fun parsePatPath ctx =
        (repeat1 (parseStrId << `DOT) && parseId wth (fn (l, x) => l @ [x]))
     || (parseNonfixId ctx wth (fn x => [x]))

    (* "=" may appear as an identifier in expressions (but not patterns)
       (technically a simplifying assumption made to parse fun decs and
       val decs correctly, but goes with the spirit of the Definition by
       preventing rebinding of "=")
       (also, see Rossberg's "Mistakes and Ambiguities" on Section 2.4) *)
    val parseExpId = parseId || `EQUALS return "="
    fun parseNonfixExpId ctx = parseExpId suchthat (isNonfix ctx)
                            || `OP >> parseExpId

    fun parseExpPath ctx =
        (repeat1 (parseStrId << `DOT) && parseExpId wth (fn (l, x) => l @ [x]))
     || (parseNonfixExpId ctx wth (fn x => [x]))

    fun parseAtomicPat_ ctx =
        alt [`UNDERSCORE return PaWild,
             parseLit wth PaLit,
             parsePatPath ctx wth PaVar,
             (* parseNonfixId ctx wth (fn x => PaVar [x]), *)
             `LPAREN >> separate0 (withCtx ctx parsePat) (`COMMA) << `RPAREN
                (* XXX check for dupes... here? *)
                wth ifmanymark paProd,
             (* XXX add support for derived form {x, y, z} -- see Defn *)
             `LBRACE >> separate0 (parseLab && `EQUALS >> withCtx ctx parsePat)
                                  (`COMMA)
                << `RBRACE
                check (uniqueLabels, fn _ => "duplicate record labels in pattern")
                wth PaRecord,
             `LBRACKET >> separate0 (withCtx ctx parsePat) (`COMMA) << `RBRACKET
                wth paList]

    and parseAtomicPat ctx = !!(withCtx ctx parseAtomicPat_)

    and parseAppPat_ ctx =
        alt [parsePatPath ctx && withCtx ctx parseAtomicPat wth PaApp,
             withCtx ctx parseAtomicPat_]

    and parseAppPat ctx = !!(withCtx ctx parseAppPat_)

    and parseInfixPat ctx =
        let fun oprItem (id, (prec, ass)) =
                    Opr (Infix (ass,
                                prec,
                                mark (fn (x, y) =>
                                        PaApp ([id], markl2 paProd [x, y]))))
            val par =
                alt [parseId when (FixContext.getInfix ctx) wth oprItem,
                     withCtx ctx parseAppPat wth Atm]
        in
            parsefixity par
        end

    and parseConstrainPat ctx =
        (*
        withCtx ctx parseInfixPat && opt (`COLON >> $parseTy)
            wth (fn (pat, SOME ty) => PaConstrain (pat, ty)
                  | (pat, NONE) => pat)
        *)
        withCtx ctx parseInfixPat && repeat (`COLON >> $parseTy)
            wth (fn (pat, tys) =>
                    foldl (fn (ty, pp) => mark PaConstrain (pp, ty)) pat tys)

    and parseAsPat ctx =
        !!(parseNonfixId ctx) && `AS >> withCtx ctx parsePat
            (* XXX check for dupes... here? *)
            wth mark (fn (idp, pat) => PaAs (forget idp, pat))

    and parsePat ctx = withCtx ctx parseAsPat || withCtx ctx parseConstrainPat 



    (*** Declarations and Expressions ***)

    fun bindable id = not (List.exists (fn i => streq i id)
                           ["true", "false", "nil", "::", "ref", "it"])

    val parseBindableId =
        parseId check (bindable,
                       fn id => "illegal rebinding of identifier " ^ id)

    fun parseBindableNonfixId ctx =
        parseBindableId suchthat (isNonfix ctx) || `OP >> parseBindableId

    fun parseBindableInfixId ctx =
        parseBindableId suchthat (FixContext.isInfix ctx)

    val parseTyVarSeq =
        alt [parseTyVar wth (fn id => [id]),
             `LPAREN >> separate parseTyVar (`COMMA) << `RPAREN
                check (uniqueStrings, fn _ => "duplicate tyvars in tyvarseq"),
             succeed []]


    (* used below in parseLocalDec and parseLocalStrDec *)
    (* NB: after we think we've found all the decs, if we can't find an `in',
       it might either be because the user meant to have more decs or the user
       forgot their `in', so we note both possibilities *)
    fun parseLocalishDec parseDecs con ctx =
      !! (
        `LOCAL >> withCtx ctx parseDecs
               << (`IN ## error "expected decs or `in'")
            -- (fn (ctx1, decs1) =>
                    withCtx ctx1 parseDecs
                    << (`END ## error "expected decs or `end'")
                    wth (fn (ctx2, decs2) =>
                            let (* add new bindings from ctx2 to outer ctx *)
                                val new = FixContext.diff ctx2 ctx1
                                val ctx = FixContext.join new ctx
                            in
                                (ctx, con (decs1, decs2))
                            end))
         )
         (* do some rearranging of the position before returning *)
         wth (fn ((ctx, dec), pos) => (ctx, (dec, pos)))

    (* used below in parseDecs, parseStrDecs, and parseTopDecs *)
    fun parseMoreDecs parseDecs ctx = fn (ctxo, deco) =>
        let val ctx = Option.getOpt (ctxo, ctx) (* possibly overrides old ctx *)
        in
            withCtx ctx parseDecs
                wth (fn (ctx, decs) => (* newly returned ctx *)
                        case deco of NONE => (ctx, decs)
                                   | SOME dec => (ctx, dec::decs))
        end


    val parseDigit = any when (fn INT i => if i >= 0 andalso i <= 9
                                            then SOME i
                                            else NONE
                                | _ => NONE)

    (* just returns a new context *)
    fun parseInfixDec ctx =
        alt [(`INFIX return Left || `INFIXR return Right)
                && optional (fn x => x) 0 parseDigit
                && repeat1 parseExpId
                wth (fn (assoc, (prec, ids)) =>
                        foldl (fn (id, ctx) =>
                                 FixContext.addInfix ctx id (prec, assoc))
                              ctx
                              ids),
             `NONFIX >> repeat1 parseExpId
                wth foldl (fn (id, ctx) => FixContext.delInfix ctx id) ctx]

    (* parametrized on parseId -- parseTyId normally, parseTyPath for
       `where type ...' clauses *)
    fun parseTybind parseId =
        parseTyVarSeq && parseId && `EQUALS >> $parseTy wth flat3

    val parseConbind = parseBindableId && opt (`OF >> $parseTy)
    val parseConbinds = parseConbind && repeat (`BAR >> parseConbind) wth op::

    val parseDatbind =
        parseTyVarSeq && parseTyId && `EQUALS >> parseConbinds wth flat3

    fun parseFunbind ctx = 
        (* first, get the (head variable, non-empty list of patterns) *)
        alt [(* fun [op] f atpat1 ... atpatn [: ty] = exp *)
             parseBindableNonfixId ctx && repeat1 (withCtx ctx parseAtomicPat),
             (* fun (atpat1 f atpat2) atpat3 ... atpatn [: ty] = exp *)
             `LPAREN >> withCtx ctx parseAtomicPat && parseBindableInfixId ctx
                && withCtx ctx parseAtomicPat << `RPAREN
                && repeat (withCtx ctx parseAtomicPat)
                wth (fn (pat1, (id, (pat2, pats))) =>
                        (id, markl2 paProd [pat1, pat2] :: pats)),
             (* fun atpat1 f atpat2 [: ty] = exp *)
             withCtx ctx parseAtomicPat && parseBindableInfixId ctx
                && withCtx ctx parseAtomicPat
                wth (fn (pat1, (id, pat2)) =>
                        (id, [markl2 paProd [pat1, pat2]]))]
        (* then the optional type constraint the body expression *)
        && opt (`COLON >> $parseTy)
        && `EQUALS >> withCtx ctx parseExp
        (* finally, return the four components flattened *)
        wth (fn ((id, pats), (tyo, body)) => (id, pats, tyo, body))

    (* Bug?  i'm not verifying that valbinds don't rebind true, false,
       nil, ::, or ref, as specified in the Definition p. 9.  i think
       the elaborator will catch those, though. *)
    and parseRegularDec ctx =
    !! (
        alt [`VAL >> (withCtx ctx parsePat ## error "expected pattern")
                && (`EQUALS ## error "expected `=' in val dec")
                >> (withCtx ctx parseExp ## error "expected expression")
                wth DecVal,
             `FUN >> (separate (withCtx ctx parseFunbind) (`AND)
                        ## error "expected `fun' clauses")
                wth DecFun,
             `TYPE >> (separate (parseTybind parseTyId) (`AND)
                        ## error "expected `type' clauses")
                wth DecType,
             `DATATYPE >> (separate parseDatbind (`AND)
                            (* check valid, "duplicate type definition",
                               "duplicate data constructor" *)
                            ## error "expected `datatype' clauses")
                && ((`WITHTYPE >> (separate (parseTybind parseTyId) (`AND)
                                    (* check valid, "duplicate type definition" *)
                                    ## error "expected `withtype' clauses"))
                    || succeed [])
                wth DecDatatype,
             `EXCEPTION >> (parseBindableId
                                ## error "expected exception identifier")
                        && opt (`OF >> ($parseTy ## error "expected type"))
                wth DecException,
             `OPEN >> (repeat1 parseStrPath ## error "expected structure names")
                wth DecOpen,
             `STRUCTURE >> separate (withCtx ctx parseStrbind) (`AND)
                wth DecStructure]
       )

    and parseLocalDec ctx = parseLocalishDec parseDecs DecLocal ctx

    (* regular decs return just a dec, infix decs just a ctx, local decs both *)
    and parseOneDec ctx =
        withCtx ctx parseRegularDec wth (fn dec => (NONE, SOME dec))
     || withCtx ctx parseInfixDec   wth (fn ctx => (SOME ctx, NONE))
     || withCtx ctx parseLocalDec   wth (fn (ctx, dec) => (SOME ctx, SOME dec))

    and parseDecs ctx = withCtx ctx parseOneDec << opt (`SEMICOLON)
                            -- parseMoreDecs parseDecs ctx
                     || succeed (ctx, nil)

    and parseAtomicExp ctx =
    !! (
        alt [parseLit wth ExLit,
             parseExpPath ctx wth ExVar,
             `HASH >> parseLab wth ExSelector,
             `LBRACE >> separate0 (parseLab && `EQUALS >> withCtx ctx parseExp)
                                  (`COMMA)
                << (`RBRACE ## error "expected record bindings, or `}'")
                check (uniqueLabels, fn _ => "duplicate record labels in exp")
                wth ExRecord,
             `LPAREN && `RPAREN return exProd [],
             (* parse parenthisized exps, sequences, and tuples all in one go.
                borrowed mostly from tom7's Hemlock parser *)
             `LPAREN >> (withCtx ctx parseExp ## error "expected exp")
                     && opt (alt [`SEMICOLON >>
                                   (separate (withCtx ctx parseExp) (`SEMICOLON)
                                    ## error "expected `;'-separated exps")
                                   wth Util.A,
                                  `COMMA >>
                                   (separate (withCtx ctx parseExp) (`COMMA)
                                    ## error "expected `,'-separated exps")
                                   wth Util.B])
                << (`RPAREN ## error "expected `)'")
                wth (fn (e, NONE) => forget e
                      | (e, SOME (Util.A es)) => exSeq (e::es)
                      | (e, SOME (Util.B es)) => exProd (e::es)),
             `LBRACKET >> separate0 (withCtx ctx parseExp) (`COMMA)
                << (`RBRACKET ## error "expected list elements, or `]'")
                wth exList,
             `LET >> withCtx ctx parseDecs
                  << (`IN ## error "expected decs or `in'")
                -- (fn (ctx, dec) => (* override ctx *)
                        !!(separate (withCtx ctx parseExp) (`SEMICOLON)
                            ## error "expected `;'-separated exps")
                        << (`END ## error "expected `end'")
                        wth (fn (es, espos) => ExLet (dec, (exSeq es, espos))))]
       )

    and parseAppExp ctx =
       (repeat1 (withCtx ctx parseAtomicExp)
        wth ListUtil.combinel (fn (ee, e) => mark ExApp (e, ee)))
    (*
    (* XXX currently broken -- should parse these as functions, really *)
    || (!!(`HASH >> parseLab) && withCtx ctx parseAtomicExp
        wth (fn (lab, exp) => (ExRecordProj (forget lab, exp),
                               Pos.union (pos lab, pos exp))))
    *)
    (* NB: now handled above in parseAtomicExp, with ExSelector instead
       of ExRecordProj *)

    and parseInfixExp ctx =
        let fun oprItem ((id, (prec, ass)), idpos) =
                    Opr (Infix (ass,
                                prec,
                                mark (fn (x, y) =>
                                        ExApp ((ExVar [id], idpos),
                                               markl2 exProd [x, y]))))
            val par =
                alt [!!(parseExpId when (FixContext.getInfix ctx)) wth oprItem,
                     withCtx ctx parseAppExp wth Atm]
        in
            parsefixity par
        end

    and parseConstrainExp ctx =
        withCtx ctx parseInfixExp
            && repeat (`COLON >> ($parseTy ## error "expected type"))
            wth (fn (exp, tys) =>
                    foldl (fn (ty, ee) => mark ExConstrain (ee, ty)) exp tys)

    and parseAndOrExp ctx =
        parsefixity
            (alt [`ANDALSO return Opr (Infix (Right, 400, mark exAndalso)),
                  `ORELSE  return Opr (Infix (Right, 300, mark exOrelse)),
                  withCtx ctx parseConstrainExp wth Atm])

    and parseMatch ctx =
       separate (withCtx ctx parsePat && `DARROW >> withCtx ctx parseExp) (`BAR)
        ## error "expected match (<pat> => <exp> | ...)"

    and parseHandleExp ctx =
        parseAndOrExp ctx && opt (`HANDLE >> !!(withCtx ctx parseMatch))
        wth (fn (exp, SOME rules) => (ExHandle (exp, forget rules),
                                      Pos.union (pos exp, pos rules))
              | (exp, NONE) => exp)

    and parseExp ctx =
    !! (
        alt [`RAISE >> (withCtx ctx parseExp ## error "expected exp")
                wth ExRaise,
             `IF >> (withCtx ctx parseExp ## error "expected `if' guard exp")
                && (`THEN ## error "expected `then'")
                    >> (withCtx ctx parseExp ## error "expected `then' exp")
                && (`ELSE ## error "expected `else'")
                    >> (withCtx ctx parseExp ## error "expected `else' exp")
                wth ExIf o flat3,
             `CASE >> (withCtx ctx parseExp ## error "expected `case' scrutinee")
                && (`OF ## error "expected `of'") >> withCtx ctx parseMatch
                wth ExCase,
             `FN >> withCtx ctx parseMatch wth ExFn]
       )
        || parseHandleExp ctx

        (* common error that can prevent us from parsing an exp: *)
        || !!(parseExpId suchthat FixContext.isInfix ctx)
            wth (fn (id, pos) => error ("misplaced infix identifier " ^ id) pos)


    (*** Signatures ***)

    and parseTydesc () =
        parseTyVarSeq && parseTyId && opt (`EQUALS >> $parseTy) wth flat3

    and parseSigExp () =
    !! (
        alt [`SIG >> $parseSpecs << `END wth SigSig,
             parseStrId wth SigVar]
       )
        &&
        repeat (`WHERE >> `TYPE >> !!(parseTybind parseTyPath))
            wth (fn (sigexp, tybinds) =>
                    foldl (mark (fn (tyb, sige) => SigWhere (sige, forget tyb)))
                          sigexp
                          tybinds)

    and parseSpec () =
    !! (
        alt [`VAL >> parseExpId && `COLON >> $parseTy wth SpecVal,
             `TYPE >> separate ($parseTydesc) (`AND) wth SpecType,
             `DATATYPE >> separate parseDatbind (`AND) wth SpecDatatype,
             `EXCEPTION >> parseBindableId && opt (`OF >> $parseTy)
                wth SpecException,
             `STRUCTURE >> separate (parseStrId && `COLON >> $parseSigExp)
                                    (`AND)
                wth SpecStructure,
             `INCLUDE >> $parseSigExp wth SpecInclude]
       )
        &&
        repeat (!! (`SHARING >> `TYPE >> parseTyPath
                        && repeat1 (`EQUALS >> parseTyPath)
                        wth op::))
            wth (fn (spec, sharings) =>
                    foldl (mark (fn (sh, sp) => SpecSharing (sp, forget sh)))
                          spec
                          sharings)

    and parseSpecs () = separate0 ($parseSpec) (opt (`SEMICOLON))


    (*** Structure-level Declarations and Structures ***)

    and parseStrbind ctx =
        parseStrId && opt ( (`COLONGREATER >> $parseSigExp) wth Util.A
                         || (`COLON >> $parseSigExp) wth Util.B )
                   && (`EQUALS >> withCtx ctx parseStrExp)
            wth (fn (strid, (NONE, str)) => (strid, str)
                  | (strid, (SOME (Util.A sg), str)) =>
                        (strid, mark StrSealOpaque (str, sg))
                  | (strid, (SOME (Util.B sg), str)) =>
                        (strid, mark StrSealTransparent (str, sg)))

    and parseStructureStrDec ctx =
     !! (
        `STRUCTURE >> separate (withCtx ctx parseStrbind) (`AND)
                      (* XXX check uniqueness *)
            wth SDecStructure
        )

    and parseLocalStrDec ctx = parseLocalishDec parseStrDecs SDecLocal ctx

    and parseOneStrDec ctx =
        withCtx ctx parseStructureStrDec
            wth (fn strdec => (NONE, SOME strdec))
     || withCtx ctx parseLocalStrDec
            wth (fn (ctx, strdec) => (SOME ctx, SOME strdec))
     || withCtx ctx parseOneDec
            wth (fn (ctxo, deco) => (ctxo, option_mapmark SDecDec deco))

    and parseStrDecs ctx = withCtx ctx parseOneStrDec << opt (`SEMICOLON)
                                -- parseMoreDecs parseStrDecs ctx
                        || succeed (ctx, nil)

    and parseStrExp ctx =
    !! (
        alt [`STRUCT >> withCtx ctx parseStrDecs
                     << (`END ## error "expected strdecs or `end'")
                wth (fn (_, decs) => StrStruct decs),
             parseStrId << `LPAREN && withCtx ctx parseStrExp << `RPAREN
                wth StrApp,
             parseStrPath
                wth StrVar,
             `LET >> withCtx ctx parseStrDecs
                  << (`IN ## error "expected strdecs or `in'")
                -- (fn (ctx, decs) =>
                        withCtx ctx parseStrExp
                        << (`END ## error "expected `end'")
                        wth (fn strexp => StrLet (decs, strexp)))]
       )
        &&
        repeat ( (`COLONGREATER >> $parseSigExp) wth Util.A
              || (`COLON >> $parseSigExp) wth Util.B )
            wth (fn (strexp, sigexps) =>
                    let fun constrain (stre, sige) =
                        case sige of
                            Util.A sige => mark StrSealOpaque (stre, sige)
                          | Util.B sige => mark StrSealTransparent (stre, sige)
                    in
                        foldl (fn (sige, stre) => constrain (stre, sige))
                              strexp
                              sigexps
                    end)



    (*** Top-level Declarations ***)

    val parseSignatureDec =
     !! (
        `SIGNATURE >> separate (parseStrId && `EQUALS >> $parseSigExp) (`AND)
            wth TDecSignature
        )

    fun parseFunctorDec ctx =
     !! (
        `FUNCTOR >> separate (parseStrId && `LPAREN >> parseStrId
                                         && `COLON >> $parseSigExp
                                         && `RPAREN >> withCtx ctx parseStrExp
                                wth flat4)
                             (`AND)
            wth TDecFunctor
        )

    fun parseOneTopDec ctx =
        parseSignatureDec
            wth (fn sigdec => (NONE, SOME sigdec))
     || withCtx ctx parseFunctorDec
            wth (fn fctdec => (NONE, SOME fctdec))
     || withCtx ctx parseOneStrDec
            wth (fn (ctxo, sdeco) => (ctxo, option_mapmark TDecStrDec sdeco))

    fun parseTopDecs ctx = withCtx ctx parseOneTopDec (* << opt (`SEMICOLON) *)
                            -- parseMoreDecs parseTopDecs ctx
                        || succeed (ctx, nil)


    fun snd (_, y) = y

    (*** for export ***)

    val parseTy = $parseTy
    val parsePat = withCtx FixContext.initial parsePat
    val parseExp = withCtx FixContext.initial parseExp
    val parseDecs = withCtx FixContext.initial parseDecs wth snd

    val parseStrExp = withCtx FixContext.initial parseStrExp
    val parseSigExp = $parseSigExp
    val parseStrDecs = withCtx FixContext.initial parseStrDecs wth snd

    val parseTopDecs = withCtx FixContext.initial parseTopDecs wth snd

    fun parse parser stream =
        case Parsing.parse parser stream of
            NONE => raise Parse (Pos.dummypos, "couldn't parse anything!")
          | SOME x => x

    (*
    fun parse parser stream =
        Stream.hd (transform parser stream)
        handle
            Parse (pos, string) =>
                (print ("Syntax error: at " ^ Pos.toString pos ^ ": " ^ string);
                 print "\n";
                 raise Fail "done")
          | Tokenize.Lex (pos, string) =>
                (print ("Lexing error: at " ^ Pos.toString pos ^ ": " ^ string);
                 print "\n";
                 raise Fail "done")
    *)

    fun parseString s =
        let val stream = StreamUtil.stostream s
            val mstream = Pos.markstream stream
            val tstream = Parsing.transform Tokenize.token mstream
        in
            parse parseTopDecs tstream
        end

    fun parseFile file =
        let val stream = StreamUtil.ftostream file
            val mstream = Pos.markstreamex file stream
            val tstream = Parsing.transform Tokenize.token mstream
        in
            parse parseTopDecs tstream
        end
  end
